@@ -3,7 +3,7 @@ package YAHC;
 use strict;
 use warnings;
 
-our $VERSION = '0.007';
+our $VERSION = '0.008';
 
 use EV;
 use Time::HiRes;
@@ -11,19 +11,19 @@ use Exporter 'import';
 use Scalar::Util qw/weaken/;
 use Fcntl qw/F_GETFL F_SETFL O_NONBLOCK/;
 use POSIX qw/EINPROGRESS EINTR EAGAIN EWOULDBLOCK strftime/;
-use Socket qw/PF_INET SOCK_STREAM $CRLF SOL_SOCKET SO_ERROR SO_RCVTIMEO SO_SNDTIMEO inet_aton inet_ntoa pack_sockaddr_in/;
+use Socket qw/PF_INET SOCK_STREAM $CRLF SOL_SOCKET SO_ERROR inet_aton inet_ntoa pack_sockaddr_in/;
 
 sub YAHC::Error::NO_ERROR                () { 0 }
 sub YAHC::Error::REQUEST_TIMEOUT         () { 1 << 0 }
 sub YAHC::Error::CONNECT_TIMEOUT         () { 1 << 1 }
-sub YAHC::Error::READ_TIMEOUT            () { 1 << 2 }
-sub YAHC::Error::WRITE_TIMEOUT           () { 1 << 3 }
-sub YAHC::Error::CONNECT_ERROR           () { 1 << 4 }
-sub YAHC::Error::READ_ERROR              () { 1 << 5 }
-sub YAHC::Error::WRITE_ERROR             () { 1 << 6 }
-sub YAHC::Error::REQUEST_ERROR           () { 1 << 7 }
-sub YAHC::Error::RESPONSE_ERROR          () { 1 << 8 }
-sub YAHC::Error::CALLBACK_ERROR          () { 1 << 9 }
+sub YAHC::Error::DRAIN_TIMEOUT           () { 1 << 2 }
+
+sub YAHC::Error::CONNECT_ERROR           () { 1 << 10 }
+sub YAHC::Error::READ_ERROR              () { 1 << 11 }
+sub YAHC::Error::WRITE_ERROR             () { 1 << 12 }
+sub YAHC::Error::REQUEST_ERROR           () { 1 << 13 }
+sub YAHC::Error::RESPONSE_ERROR          () { 1 << 14 }
+sub YAHC::Error::CALLBACK_ERROR          () { 1 << 15 }
 sub YAHC::Error::INTERNAL_ERROR          () { 1 << 31 }
 
 sub YAHC::State::INITIALIZED             () { 1 << 0 }
@@ -89,8 +89,8 @@ sub new {
         connections         => $storage{connections},
         last_connection_id  => $$ * 1000,
 
-        debug               => delete $args->{debug},
-        keep_timeline       => delete $args->{keep_timeline},
+        debug               => delete $args->{debug} || 0,
+        keep_timeline       => delete $args->{keep_timeline} || 0,
         pool_args           => $args,
     }, $class;
 
@@ -148,7 +148,6 @@ sub request {
     $self->{callbacks}{$conn_id} = \%callbacks;
     $self->{connections}{$conn_id} = $conn;
 
-    $self->_set_request_timer($conn_id) if $conn->{request}{request_timeout};
     $self->_set_init_state($conn_id);
     return $conn;
 }
@@ -315,13 +314,21 @@ sub _set_init_state {
             return;
         }
 
+        # If later I would need timeout covering multiple attempts
+        # I can set them here under $attempt == 0 condition.
+        # Also, in _set_until_state_timer I would go to USER_ACTION_STATE
+        # instead of reinitializing connection.
+
+        $self->_set_request_timer($conn_id)    if $conn->{request}{request_timeout};
+        $self->_set_connection_timer($conn_id) if $conn->{request}{connect_timeout};
+        $self->_set_drain_timer($conn_id)      if $conn->{request}{drain_timeout};
+
         eval {
             my ($host, $ip, $port) = _get_next_target($conn);
             _register_in_timeline($conn, "Target $host:$port ($ip:$port) chosen for attempt #$attempt") if $conn->{keep_timeline};
 
             my $sock = _build_socket_and_connect($ip, $port, $conn->{request});
             $self->_set_wait_synack_state($conn_id, $sock, $ip, $host, $port);
-            $self->_set_connection_timer($conn_id) if $conn->{request}{connect_timeout};
             $continue = 0;
             1;
         } or do {
@@ -504,6 +511,8 @@ sub _set_user_action_state {
         return;
     };
 
+    $self->{loop}->now_update;
+
     my $state = $conn->{state};
     _register_in_timeline($conn, "after invoking callback state is %s", _strstate($state)) if $conn->{keep_timeline};
 
@@ -561,18 +570,6 @@ sub _build_socket_and_connect {
     my $flags = fcntl($sock, F_GETFL, 0) or die "YAHC: Failed to get fcntl F_GETFL flag: $!\n";
     fcntl($sock, F_SETFL, $flags | O_NONBLOCK) or die "YAHC: Failed to set fcntl O_NONBLOCK flag: $!\n";
 
-    if (my $read_timeout = $timeouts->{read_timeout}) {
-        my $read_struct  = pack('l!l!', $read_timeout, 0);
-        setsockopt($sock, SOL_SOCKET, SO_RCVTIMEO, $read_struct)
-          or die "YAHC: Failed to set setsockopt(SO_RCVTIMEO): $!\n";
-    }
-
-    if (my $write_timeout = $timeouts->{write_timeout}) {
-        my $write_struct = pack('l!l!', $write_timeout, 0);
-        setsockopt($sock, SOL_SOCKET, SO_SNDTIMEO, $write_struct )
-          or die "YAHC: Failed to set setsockopt(SO_SNDTIMEO): $!\n";
-    }
-
     my $ip_addr = inet_aton($ip);
     my $addr = pack_sockaddr_in($port, $ip_addr);
     if (!connect($sock, $addr) && $! != EINPROGRESS) {
@@ -599,60 +596,37 @@ sub _get_next_target {
 # Timers
 ################################################################################
 
-sub _set_request_timer {
-    my ($self, $conn_id) = @_;
+sub _set_request_timer    { $_[0]->_set_until_state_timer($_[1], 'request_timeout', YAHC::State::USER_ACTION(), YAHC::Error::REQUEST_TIMEOUT()) }
+sub _set_connection_timer { $_[0]->_set_until_state_timer($_[1], 'connect_timeout', YAHC::State::CONNECTED(),   YAHC::Error::CONNECT_TIMEOUT()) }
+sub _set_drain_timer      { $_[0]->_set_until_state_timer($_[1], 'drain_timeout',   YAHC::State::READING(),     YAHC::Error::DRAIN_TIMEOUT())   }
 
+sub _set_until_state_timer {
+    my ($self, $conn_id, $timeout_name, $state, $error_to_report) = @_;
+
+    my $timer_name = $timeout_name . '_timer';
     my $conn = $self->{connections}{$conn_id};
     my $watchers = $self->{watchers}{$conn_id};
     die "YAHC: unknown connection id $conn_id\n" unless $conn && $watchers;
 
-    if (my $timer = delete $watchers->{connection_timer}) {
+    if (my $timer = delete $watchers->{$timer_name}) {
         $timer->stop;
     }
 
-    my $timeout = $conn->{request}{request_timeout};
-    return unless defined $timeout;
-
-    my $request_timer_cb = $self->_get_safe_wrapper($conn, sub {
-        if ($conn->{state} < YAHC::State::USER_ACTION()) {
-            my $error = sprintf("Request timeout of %.3fs has reached", $timeout);
-            $self->_set_user_action_state($conn_id, YAHC::Error::REQUEST_TIMEOUT(), $error);
-        } else {
-            _register_in_timeline($conn, "delete request timer") if $conn->{keep_timeline};
-        }
-    });
-
-    _register_in_timeline($conn, "setting request timeout to %.3fs", $timeout) if $conn->{keep_timeline};
-    $watchers->{request_timer} = $self->{loop}->timer($timeout, 0, $request_timer_cb);
-    $watchers->{request_timer}->priority(2); # set max priority
-}
-
-sub _set_connection_timer {
-    my ($self, $conn_id) = @_;
-
-    my $conn = $self->{connections}{$conn_id};
-    my $watchers = $self->{watchers}{$conn_id};
-    die "YAHC: unknown connection id $conn_id\n" unless $conn && $watchers;
-
-    if (my $timer = delete $watchers->{connection_timer}) {
-        $timer->stop;
-    }
-
-    my $timeout = $conn->{request}{connect_timeout};
+    my $timeout = $conn->{request}{$timeout_name};
     return unless $timeout;
 
-    my $connection_timer_cb = $self->_get_safe_wrapper($conn, sub {
-        if ($conn->{state} < YAHC::State::CONNECTED()) {
-            _register_in_timeline($conn, "connection timeout of %.3fs has reached", $timeout) if $conn->{keep_timeline};
+    my $timer_cb = $self->_get_safe_wrapper($conn, sub {
+        if ($conn->{state} < $state) {
+            _register_error($conn, $error_to_report, "$timeout_name of %.3fs expired", $timeout);
             $self->_set_init_state($conn_id);
         } else {
-            _register_in_timeline($conn, "delete connection timer") if $conn->{keep_timeline};
+            _register_in_timeline($conn, "delete $timer_name") if $conn->{keep_timeline};
         }
     });
 
-    _register_in_timeline($conn, "setting connection timeout to %.3fs", $timeout) if $conn->{keep_timeline};
-    $watchers->{connection_timer} = $self->{loop}->timer($timeout, 0, $connection_timer_cb);
-    $watchers->{connection_timer}->priority(-2); # set lowest priority
+    _register_in_timeline($conn, "setting $timeout_name to %.3fs", $timeout) if $conn->{keep_timeline};
+    $watchers->{$timer_name} = $self->{loop}->timer($timeout, 0, $timer_cb);
+    $watchers->{$timer_name}->priority(-2); # set lowest priority
 }
 
 ################################################################################
@@ -731,6 +705,8 @@ sub _call_state_callback {
         warn "YAHC: exception in callback: $error";
         $self->_set_completed_state($conn->{id});
     };
+
+    # $self->{loop}->now_update; # XXX expect state callbacks to be small
 }
 
 sub _get_safe_wrapper {
